@@ -69,11 +69,19 @@ std::shared_ptr<HeifFile> ImageItem::get_file() const
 
 heif_property_id ImageItem::add_property(std::shared_ptr<Box> property, bool essential)
 {
+  // TODO: is this correct? What happens when add_property does deduplicate the property?
   m_properties.push_back(property);
   return get_file()->add_property(get_id(), property, essential);
 }
 
 
+heif_property_id ImageItem::add_property_without_deduplication(std::shared_ptr<Box> property, bool essential)
+{
+  m_properties.push_back(property);
+  return get_file()->add_property_without_deduplication(get_id(), property, essential);
+}
+
+
 Error ImageItem::init_decoder_from_item(heif_item_id id)
 {
   m_id = id;
@@ -596,26 +604,27 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::convert_colorspace_for_encodi
 
   std::shared_ptr<HeifPixelImage> output_image;
 
-  if (colorspace != image->get_colorspace() ||
-      chroma != image->get_chroma_format() ||
-      !nclx_profile_matches_spec(colorspace, image->get_color_profile_nclx(), output_nclx_profile)) {
-    // @TODO: use color profile when converting
-    int output_bpp = 0; // same as input
-
-    //auto target_nclx = std::make_shared<color_profile_nclx>();
-    //target_nclx->set_from_heif_color_profile_nclx(target_heif_nclx);
-
-    output_image = convert_colorspace(image, colorspace, chroma, target_nclx_profile,
-                                      output_bpp, options.color_conversion_options);
-    if (!output_image) {
-      return Error(heif_error_Unsupported_feature, heif_suberror_Unsupported_color_conversion);
-    }
-  }
-  else {
-    output_image = image;
+  if (colorspace == image->get_colorspace() &&
+      chroma == image->get_chroma_format() &&
+      nclx_profile_matches_spec(colorspace, image->get_color_profile_nclx(), output_nclx_profile)) {
+    return image;
   }
 
-  return output_image;
+
+  // @TODO: use color profile when converting
+  int output_bpp = 0; // same as input
+
+  //auto target_nclx = std::make_shared<color_profile_nclx>();
+  //target_nclx->set_from_heif_color_profile_nclx(target_heif_nclx);
+
+  auto output_image_result = convert_colorspace(image, colorspace, chroma, target_nclx_profile,
+                                                output_bpp, options.color_conversion_options,
+                                                get_context()->get_security_limits());
+  if (output_image_result.error) {
+    return output_image_result.error;
+  }
+
+  return *output_image_result;
 }
 
 
@@ -791,7 +800,7 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
 
     for (const auto& property : properties) {
       if (auto rot = std::dynamic_pointer_cast<Box_irot>(property)) {
-        auto rotateResult = img->rotate_ccw(rot->get_rotation_ccw());
+        auto rotateResult = img->rotate_ccw(rot->get_rotation_ccw(), m_heif_context->get_security_limits());
         if (rotateResult.error) {
           return error;
         }
@@ -801,7 +810,8 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
 
 
       if (auto mirror = std::dynamic_pointer_cast<Box_imir>(property)) {
-        auto mirrorResult = img->mirror_inplace(mirror->get_mirror_direction());
+        auto mirrorResult = img->mirror_inplace(mirror->get_mirror_direction(),
+                                                get_context()->get_security_limits());
         if (mirrorResult.error) {
           return error;
         }
@@ -835,7 +845,7 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
                          heif_suberror_Invalid_clean_aperture);
           }
 
-          auto cropResult = img->crop(left, right, top, bottom);
+          auto cropResult = img->crop(left, right, top, bottom, m_heif_context->get_security_limits());
           if (error) {
             return error;
           }
@@ -890,7 +900,7 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
 
     if ((alpha_image->get_width() != img->get_width()) || (alpha_image->get_height() != img->get_height())) {
       std::shared_ptr<HeifPixelImage> scaled_alpha;
-      Error err = alpha->scale_nearest_neighbor(scaled_alpha, img->get_width(), img->get_height());
+      Error err = alpha->scale_nearest_neighbor(scaled_alpha, img->get_width(), img->get_height(), m_heif_context->get_security_limits());
       if (err) {
         return err;
       }
