@@ -189,7 +189,8 @@ static uint32_t rounded_size(uint32_t s)
   return s;
 }
 
-bool HeifPixelImage::add_plane(heif_channel channel, uint32_t width, uint32_t height, int bit_depth)
+Error HeifPixelImage::add_plane(heif_channel channel, uint32_t width, uint32_t height, int bit_depth,
+                                const heif_security_limits* limits)
 {
   assert(!has_channel(channel));
 
@@ -206,31 +207,33 @@ bool HeifPixelImage::add_plane(heif_channel channel, uint32_t width, uint32_t he
     bit_depth = 8;
   }
 
-  if (plane.alloc(width, height, heif_channel_datatype_unsigned_integer, bit_depth, num_interleaved_pixels)) {
-    m_planes.insert(std::make_pair(channel, plane));
-    return true;
+  if (auto err = plane.alloc(width, height, heif_channel_datatype_unsigned_integer, bit_depth, num_interleaved_pixels, limits)) {
+    return err;
   }
   else {
-    return false;
+    m_planes.insert(std::make_pair(channel, plane));
+    return Error::Ok;
   }
 }
 
 
-bool HeifPixelImage::add_channel(heif_channel channel, uint32_t width, uint32_t height, heif_channel_datatype datatype, int bit_depth)
+Error HeifPixelImage::add_channel(heif_channel channel, uint32_t width, uint32_t height, heif_channel_datatype datatype, int bit_depth,
+                                  const heif_security_limits* limits)
 {
   ImagePlane plane;
-  if (plane.alloc(width, height, datatype, bit_depth, 1)) {
-    m_planes.insert(std::make_pair(channel, plane));
-    return true;
+  if (Error err = plane.alloc(width, height, datatype, bit_depth, 1, limits)) {
+    return err;
   }
   else {
-    return false;
+    m_planes.insert(std::make_pair(channel, plane));
+    return Error::Ok;
   }
 }
 
 
-bool HeifPixelImage::ImagePlane::alloc(uint32_t width, uint32_t height, heif_channel_datatype datatype, int bit_depth,
-                                       int num_interleaved_components) // heif_chroma chroma)
+Error HeifPixelImage::ImagePlane::alloc(uint32_t width, uint32_t height, heif_channel_datatype datatype, int bit_depth,
+                                        int num_interleaved_components,
+                                        const heif_security_limits* limits)
 {
   assert(bit_depth >= 1);
   assert(bit_depth <= 128);
@@ -257,8 +260,19 @@ bool HeifPixelImage::ImagePlane::alloc(uint32_t width, uint32_t height, heif_cha
   stride = m_mem_width * bytes_per_pixel;
   stride = (stride + alignment - 1U) & ~(alignment - 1U);
 
-  if ((heif_get_global_security_limits()->max_memory_block_size - (alignment + 1)) / stride < m_mem_height) {
-    return false;
+  assert(alignment>=1);
+
+  if (limits &&
+      limits->max_memory_block_size &&
+      (limits->max_memory_block_size < alignment - 1U ||
+       (limits->max_memory_block_size - (alignment - 1U)) / stride < m_mem_height)) {
+    std::stringstream sstr;
+    sstr << "Allocating " << static_cast<size_t>(m_mem_height) * stride + alignment - 1 << " exceeds the security limit of "
+         << limits->max_memory_block_size << " bytes";
+
+    return {heif_error_Memory_allocation_error,
+            heif_suberror_Security_limit_exceeded,
+            sstr.str()};
   }
 
   try {
@@ -275,15 +289,21 @@ bool HeifPixelImage::ImagePlane::alloc(uint32_t width, uint32_t height, heif_cha
 
     mem = mem_8;
 
-    return true;
+    return Error::Ok;
   }
   catch (const std::bad_alloc& excpt) {
-    return false;
+    std::stringstream sstr;
+    sstr << "Allocating " << static_cast<size_t>(m_mem_height) * stride + alignment - 1 << " bytes failed";
+
+    return {heif_error_Memory_allocation_error,
+            heif_suberror_Unspecified,
+            sstr.str()};
   }
 }
 
 
-bool HeifPixelImage::extend_padding_to_size(uint32_t width, uint32_t height, bool adjust_size)
+Error HeifPixelImage::extend_padding_to_size(uint32_t width, uint32_t height, bool adjust_size,
+                                             const heif_security_limits* limits)
 {
   for (auto& planeIter : m_planes) {
     auto* plane = &planeIter.second;
@@ -301,8 +321,16 @@ bool HeifPixelImage::extend_padding_to_size(uint32_t width, uint32_t height, boo
         plane->m_mem_height < subsampled_height) {
 
       ImagePlane newPlane;
-      if (!newPlane.alloc(subsampled_width, subsampled_height, plane->m_datatype, plane->m_bit_depth, num_interleaved_pixels_per_plane(m_chroma))) {
-        return false;
+      if (auto err = newPlane.alloc(subsampled_width, subsampled_height, plane->m_datatype, plane->m_bit_depth,
+                                    num_interleaved_pixels_per_plane(m_chroma),
+                                    limits))
+      {
+        return err;
+      }
+
+      // This is not needed, but we have to silence the clang-tidy false positive.
+      if (!newPlane.mem) {
+        return Error::InternalError;
       }
 
       // copy the visible part of the old plane into the new plane
@@ -349,11 +377,11 @@ bool HeifPixelImage::extend_padding_to_size(uint32_t width, uint32_t height, boo
     m_height = height;
   }
 
-  return true;
+  return Error::Ok;
 }
 
 
-bool HeifPixelImage::extend_to_size_with_zero(uint32_t width, uint32_t height)
+Error HeifPixelImage::extend_to_size_with_zero(uint32_t width, uint32_t height, const heif_security_limits* limits)
 {
   for (auto& planeIter : m_planes) {
     auto* plane = &planeIter.second;
@@ -371,8 +399,13 @@ bool HeifPixelImage::extend_to_size_with_zero(uint32_t width, uint32_t height)
         plane->m_mem_height < subsampled_height) {
 
       ImagePlane newPlane;
-      if (!newPlane.alloc(subsampled_width, subsampled_height, plane->m_datatype, plane->m_bit_depth, num_interleaved_pixels_per_plane(m_chroma))) {
-        return false;
+      if (auto err = newPlane.alloc(subsampled_width, subsampled_height, plane->m_datatype, plane->m_bit_depth, num_interleaved_pixels_per_plane(m_chroma), limits)) {
+        return err;
+      }
+
+      // This is not needed, but we have to silence the clang-tidy false positive.
+      if (!newPlane.mem) {
+        return Error::InternalError;
       }
 
       // copy the visible part of the old plane into the new plane
@@ -413,7 +446,7 @@ bool HeifPixelImage::extend_to_size_with_zero(uint32_t width, uint32_t height)
   m_width = width;
   m_height = height;
 
-  return true;
+  return Error::Ok;
 }
 
 bool HeifPixelImage::has_channel(heif_channel channel) const
@@ -525,9 +558,10 @@ int HeifPixelImage::get_number_of_interleaved_components(heif_channel channel) c
 }
 
 
-void HeifPixelImage::copy_new_plane_from(const std::shared_ptr<const HeifPixelImage>& src_image,
-                                         heif_channel src_channel,
-                                         heif_channel dst_channel)
+Error HeifPixelImage::copy_new_plane_from(const std::shared_ptr<const HeifPixelImage>& src_image,
+                                          heif_channel src_channel,
+                                          heif_channel dst_channel,
+                                          const heif_security_limits* limits)
 {
   assert(src_image->has_channel(src_channel));
   assert(!has_channel(dst_channel));
@@ -539,9 +573,12 @@ void HeifPixelImage::copy_new_plane_from(const std::shared_ptr<const HeifPixelIm
   assert(src_plane_iter != src_image->m_planes.end());
   const auto& src_plane = src_plane_iter->second;
 
-  add_channel(dst_channel, width, height,
-              src_plane.m_datatype,
-              src_image->get_bits_per_pixel(src_channel));
+  auto err = add_channel(dst_channel, width, height,
+                         src_plane.m_datatype,
+                         src_image->get_bits_per_pixel(src_channel), limits);
+  if (err) {
+    return err;
+  }
 
   uint8_t* dst;
   uint32_t dst_stride = 0;
@@ -557,15 +594,20 @@ void HeifPixelImage::copy_new_plane_from(const std::shared_ptr<const HeifPixelIm
   for (uint32_t y = 0; y < height; y++) {
     memcpy(dst + y * dst_stride, src + y * src_stride, bpl);
   }
+
+  return Error::Ok;
 }
 
 
-void HeifPixelImage::extract_alpha_from_RGBA(const std::shared_ptr<const HeifPixelImage>& src_image)
+Error HeifPixelImage::extract_alpha_from_RGBA(const std::shared_ptr<const HeifPixelImage>& src_image,
+                                              const heif_security_limits* limits)
 {
   uint32_t width = src_image->get_width();
   uint32_t height = src_image->get_height();
 
-  add_plane(heif_channel_Y, width, height, src_image->get_bits_per_pixel(heif_channel_interleaved));
+  if (Error err = add_plane(heif_channel_Y, width, height, src_image->get_bits_per_pixel(heif_channel_interleaved), limits)) {
+    return err;
+  }
 
   uint8_t* dst;
   uint32_t dst_stride = 0;
@@ -583,13 +625,21 @@ void HeifPixelImage::extract_alpha_from_RGBA(const std::shared_ptr<const HeifPix
       dst[y * dst_stride + x] = src[y * src_stride + 4 * x + 3];
     }
   }
+
+  return Error::Ok;
 }
 
 
-void HeifPixelImage::fill_new_plane(heif_channel dst_channel, uint16_t value, int width, int height, int bpp)
+Error HeifPixelImage::fill_new_plane(heif_channel dst_channel, uint16_t value, int width, int height, int bpp,
+                                     const heif_security_limits* limits)
 {
-  add_plane(dst_channel, width, height, bpp);
+  if (Error err = add_plane(dst_channel, width, height, bpp, limits)) {
+    return err;
+  }
+
   fill_plane(dst_channel, value);
+
+  return Error::Ok;
 }
 
 
@@ -714,7 +764,7 @@ Error HeifPixelImage::copy_image_to(const std::shared_ptr<const HeifPixelImage>&
 }
 
 
-Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::rotate_ccw(int angle_degrees)
+Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::rotate_ccw(int angle_degrees, const heif_security_limits* limits)
 {
   // --- for some subsampled chroma colorspaces, we have to transform to 4:4:4 before rotation
 
@@ -744,8 +794,12 @@ Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::rotate_ccw(int angle_deg
     heif_color_conversion_options options{};
     heif_color_conversion_options_set_defaults(&options);
 
-    auto converted_image = convert_colorspace(shared_from_this(), heif_colorspace_YCbCr, heif_chroma_444, nullptr, get_bits_per_pixel(heif_channel_Y), options);
-    return converted_image->rotate_ccw(angle_degrees);
+    auto converted_image_result = convert_colorspace(shared_from_this(), heif_colorspace_YCbCr, heif_chroma_444, nullptr, get_bits_per_pixel(heif_channel_Y), options, limits);
+    if (converted_image_result.error) {
+      return converted_image_result.error;
+    }
+
+    return (*converted_image_result)->rotate_ccw(angle_degrees, limits);
   }
 
 
@@ -787,7 +841,10 @@ Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::rotate_ccw(int angle_deg
       std::swap(out_plane_width, out_plane_height);
     }
 
-    out_img->add_channel(channel, out_plane_width, out_plane_height, plane.m_datatype, plane.m_bit_depth);
+    Error err = out_img->add_channel(channel, out_plane_width, out_plane_height, plane.m_datatype, plane.m_bit_depth, limits);
+    if (err) {
+      return err;
+    }
 
     auto out_plane_iter = out_img->m_planes.find(channel);
     assert(out_plane_iter != out_img->m_planes.end());
@@ -871,7 +928,8 @@ void HeifPixelImage::ImagePlane::mirror_inplace(heif_transform_mirror_direction
 }
 
 
-Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::mirror_inplace(heif_transform_mirror_direction direction)
+Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::mirror_inplace(heif_transform_mirror_direction direction,
+                                                                       const heif_security_limits* limits)
 {
   // --- for some subsampled chroma colorspaces, we have to transform to 4:4:4 before rotation
 
@@ -892,8 +950,12 @@ Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::mirror_inplace(heif_tran
     heif_color_conversion_options options{};
     heif_color_conversion_options_set_defaults(&options);
 
-    auto converted_image = convert_colorspace(shared_from_this(), heif_colorspace_YCbCr, heif_chroma_444, nullptr, get_bits_per_pixel(heif_channel_Y), options);
-    return converted_image->mirror_inplace(direction);
+    auto converted_image_result = convert_colorspace(shared_from_this(), heif_colorspace_YCbCr, heif_chroma_444, nullptr, get_bits_per_pixel(heif_channel_Y), options, limits);
+    if (converted_image_result.error) {
+      return converted_image_result.error;
+    }
+
+    return (*converted_image_result)->mirror_inplace(direction, limits);
   }
 
 
@@ -949,7 +1011,8 @@ int HeifPixelImage::ImagePlane::get_bytes_per_pixel() const
 }
 
 
-Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::crop(uint32_t left, uint32_t right, uint32_t top, uint32_t bottom) const
+Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::crop(uint32_t left, uint32_t right, uint32_t top, uint32_t bottom,
+                                                             const heif_security_limits* limits) const
 {
   // --- for some subsampled chroma colorspaces, we have to transform to 4:4:4 before rotation
 
@@ -967,8 +1030,12 @@ Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::crop(uint32_t left, uint
     heif_color_conversion_options options{};
     heif_color_conversion_options_set_defaults(&options);
 
-    auto converted_image = convert_colorspace(shared_from_this(), heif_colorspace_YCbCr, heif_chroma_444, nullptr, get_bits_per_pixel(heif_channel_Y), options);
-    return converted_image->crop(left, right, top, bottom);
+    auto converted_image_result = convert_colorspace(shared_from_this(), heif_colorspace_YCbCr, heif_chroma_444, nullptr, get_bits_per_pixel(heif_channel_Y), options, limits);
+    if (converted_image_result.error) {
+      return converted_image_result.error;
+    }
+
+    return (*converted_image_result)->crop(left, right, top, bottom, limits);
   }
 
 
@@ -991,11 +1058,15 @@ Result<std::shared_ptr<HeifPixelImage>> HeifPixelImage::crop(uint32_t left, uint
     uint32_t plane_top = top * h / m_height;
     uint32_t plane_bottom = bottom * h / m_height;
 
-    out_img->add_channel(channel,
-                         plane_right - plane_left + 1,
-                         plane_bottom - plane_top + 1,
-                         plane.m_datatype,
-                         plane.m_bit_depth);
+    auto err = out_img->add_channel(channel,
+                                    plane_right - plane_left + 1,
+                                    plane_bottom - plane_top + 1,
+                                    plane.m_datatype,
+                                    plane.m_bit_depth,
+                                    limits);
+    if (err) {
+      return err;
+    }
 
     auto out_plane_iter = out_img->m_planes.find(channel);
     assert(out_plane_iter != out_img->m_planes.end());
@@ -1241,7 +1312,8 @@ Error HeifPixelImage::overlay(std::shared_ptr<HeifPixelImage>& overlay, int32_t
 
 
 Error HeifPixelImage::scale_nearest_neighbor(std::shared_ptr<HeifPixelImage>& out_img,
-                                             uint32_t width, uint32_t height) const
+                                             uint32_t width, uint32_t height,
+                                             const heif_security_limits* limits) const
 {
   out_img = std::make_shared<HeifPixelImage>();
   out_img->create(width, height, m_colorspace, m_chroma);
@@ -1250,7 +1322,9 @@ Error HeifPixelImage::scale_nearest_neighbor(std::shared_ptr<HeifPixelImage>& ou
   // --- create output image with scaled planes
 
   if (has_channel(heif_channel_interleaved)) {
-    out_img->add_plane(heif_channel_interleaved, width, height, get_bits_per_pixel(heif_channel_interleaved));
+    if (auto err = out_img->add_plane(heif_channel_interleaved, width, height, get_bits_per_pixel(heif_channel_interleaved), limits)) {
+      return err;
+    }
   }
   else {
     if (get_colorspace() == heif_colorspace_RGB) {
@@ -1260,16 +1334,24 @@ Error HeifPixelImage::scale_nearest_neighbor(std::shared_ptr<HeifPixelImage>& ou
         return {heif_error_Invalid_input, heif_suberror_Unspecified, "RGB input without R,G,B, planes"};
       }
 
-      out_img->add_plane(heif_channel_R, width, height, get_bits_per_pixel(heif_channel_R));
-      out_img->add_plane(heif_channel_G, width, height, get_bits_per_pixel(heif_channel_G));
-      out_img->add_plane(heif_channel_B, width, height, get_bits_per_pixel(heif_channel_B));
+      if (auto err = out_img->add_plane(heif_channel_R, width, height, get_bits_per_pixel(heif_channel_R), limits)) {
+        return err;
+      }
+      if (auto err = out_img->add_plane(heif_channel_G, width, height, get_bits_per_pixel(heif_channel_G), limits)) {
+        return err;
+      }
+      if (auto err = out_img->add_plane(heif_channel_B, width, height, get_bits_per_pixel(heif_channel_B), limits)) {
+        return err;
+      }
     }
     else if (get_colorspace() == heif_colorspace_monochrome) {
       if (!has_channel(heif_channel_Y)) {
         return {heif_error_Invalid_input, heif_suberror_Unspecified, "monochrome input with no Y plane"};
       }
 
-      out_img->add_plane(heif_channel_Y, width, height, get_bits_per_pixel(heif_channel_Y));
+      if (auto err = out_img->add_plane(heif_channel_Y, width, height, get_bits_per_pixel(heif_channel_Y), limits)) {
+        return err;
+      }
     }
     else if (get_colorspace() == heif_colorspace_YCbCr) {
       if (!has_channel(heif_channel_Y) ||
@@ -1280,16 +1362,24 @@ Error HeifPixelImage::scale_nearest_neighbor(std::shared_ptr<HeifPixelImage>& ou
 
       uint32_t cw, ch;
       get_subsampled_size(width, height, heif_channel_Cb, get_chroma_format(), &cw, &ch);
-      out_img->add_plane(heif_channel_Y, width, height, get_bits_per_pixel(heif_channel_Y));
-      out_img->add_plane(heif_channel_Cb, cw, ch, get_bits_per_pixel(heif_channel_Cb));
-      out_img->add_plane(heif_channel_Cr, cw, ch, get_bits_per_pixel(heif_channel_Cr));
+      if (auto err = out_img->add_plane(heif_channel_Y, width, height, get_bits_per_pixel(heif_channel_Y), limits)) {
+        return err;
+      }
+      if (auto err = out_img->add_plane(heif_channel_Cb, cw, ch, get_bits_per_pixel(heif_channel_Cb), limits)) {
+        return err;
+      }
+      if (auto err = out_img->add_plane(heif_channel_Cr, cw, ch, get_bits_per_pixel(heif_channel_Cr), limits)) {
+        return err;
+      }
     }
     else {
       return {heif_error_Invalid_input, heif_suberror_Unspecified, "unknown color configuration"};
     }
 
     if (has_channel(heif_channel_Alpha)) {
-      out_img->add_plane(heif_channel_Alpha, width, height, get_bits_per_pixel(heif_channel_Alpha));
+      if (auto err = out_img->add_plane(heif_channel_Alpha, width, height, get_bits_per_pixel(heif_channel_Alpha), limits)) {
+        return err;
+      }
     }
   }
 
@@ -1365,7 +1455,8 @@ void HeifPixelImage::debug_dump() const
   }
 }
 
-void HeifPixelImage::create_clone_image_at_new_size(const std::shared_ptr<const HeifPixelImage>& source, uint32_t w, uint32_t h)
+Error HeifPixelImage::create_clone_image_at_new_size(const std::shared_ptr<const HeifPixelImage>& source, uint32_t w, uint32_t h,
+                                                     const heif_security_limits* limits)
 {
   heif_colorspace colorspace = source->get_colorspace();
   heif_chroma chroma = source->get_chroma_format();
@@ -1374,17 +1465,31 @@ void HeifPixelImage::create_clone_image_at_new_size(const std::shared_ptr<const
 
   switch (colorspace) {
     case heif_colorspace_monochrome:
-      add_plane(heif_channel_Y, w, h, source->get_bits_per_pixel(heif_channel_Y));
+      if (auto err = add_plane(heif_channel_Y, w, h, source->get_bits_per_pixel(heif_channel_Y), limits)) {
+        return err;
+      }
       break;
     case heif_colorspace_YCbCr:
-      add_plane(heif_channel_Y, w, h, source->get_bits_per_pixel(heif_channel_Y));
-      add_plane(heif_channel_Cb, chroma_width(w, chroma), chroma_height(h, chroma), source->get_bits_per_pixel(heif_channel_Cb));
-      add_plane(heif_channel_Cr, chroma_width(w, chroma), chroma_height(h, chroma), source->get_bits_per_pixel(heif_channel_Cr));
+      if (auto err = add_plane(heif_channel_Y, w, h, source->get_bits_per_pixel(heif_channel_Y), limits)) {
+        return err;
+      }
+      if (auto err = add_plane(heif_channel_Cb, chroma_width(w, chroma), chroma_height(h, chroma), source->get_bits_per_pixel(heif_channel_Cb), limits)) {
+        return err;
+      }
+      if (auto err = add_plane(heif_channel_Cr, chroma_width(w, chroma), chroma_height(h, chroma), source->get_bits_per_pixel(heif_channel_Cr), limits)) {
+        return err;
+      }
       break;
     case heif_colorspace_RGB:
-      add_plane(heif_channel_R, w, h, source->get_bits_per_pixel(heif_channel_R));
-      add_plane(heif_channel_G, w, h, source->get_bits_per_pixel(heif_channel_G));
-      add_plane(heif_channel_B, w, h, source->get_bits_per_pixel(heif_channel_B));
+      if (auto err = add_plane(heif_channel_R, w, h, source->get_bits_per_pixel(heif_channel_R), limits)) {
+        return err;
+      }
+      if (auto err = add_plane(heif_channel_G, w, h, source->get_bits_per_pixel(heif_channel_G), limits)) {
+        return err;
+      }
+      if (auto err = add_plane(heif_channel_B, w, h, source->get_bits_per_pixel(heif_channel_B), limits)) {
+        return err;
+      }
       break;
     default:
       assert(false);
@@ -1392,6 +1497,10 @@ void HeifPixelImage::create_clone_image_at_new_size(const std::shared_ptr<const
   }
 
   if (source->has_alpha()) {
-      add_plane(heif_channel_Alpha, w, h, source->get_bits_per_pixel(heif_channel_Alpha));
+      if (auto err = add_plane(heif_channel_Alpha, w, h, source->get_bits_per_pixel(heif_channel_Alpha), limits)) {
+        return err;
+      }
   }
+
+  return Error::Ok;
 }
